/*

  Nosuid TCP/IP ping 1.5 beta by Michal Zalewski <lcamtuf@ids.pl>
  ---------------------------------------------------------------
 
  The Nosuid TCP/IP ping and related utilities are free software; you
  can redistribute it and/or modify it under the terms of the GNU Library
  General Public License as published by the Free Software Foundation;
  either version 2 of the License, or (at your option) any later version.
 
*/

#include <stdio.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdarg.h>
#include <errno.h>
#include <signal.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>

struct hostent *hp;
struct sockaddr_in s;
int suck,nic,to,tamto,count,delay=1,timeout=5,n=1,got,min,max,trip,skew,port;
struct timeval tv,tv2;
char ad[20],ch;
struct sigaction siga;


void xerr(char* when) {
  perror(when);
  exit(errno);
}


void usage(void) {
  fprintf(stderr,"  usage: ping [ -i delay ] [ -c count ] [ -t timeout] host\n");
  exit(2);
}


void alrm(int x) {to=1;}


void stats(int x) {
  printf("\n-- %s ping statistics --\n",ad);
  printf("%d packet(s) transmitted, %d packet(s) received, %d%% packets lost.\n",n-1,got,(n-1-got)*100/(n-1));
  printf("round-trip statistics min/avg/max: ");
  if (got) {
     printf("%.1f/%.1f/%.1f ms",(float)min/1000,(float)trip/got/1000,(float)max/1000);
     if (min/1000<=0.1) printf(" (%d/%d/%d usecs)",min,trip/got,max);
     printf(".\n");
  } else printf("(not available due to 100%% packets lost).\n");
  exit(0);
}


int main(int argc,char* argv[]) {
  
  printf("Nosuid TCP/IP ping 1.5 beta by <lcamtuf@ids.pl>\n");

  while ((ch=getopt(argc,argv,"i:t:c:")) != EOF)
    switch(ch) {
      case 'c':
        count=atoi(optarg);
        if (count<0) { fprintf(stderr,"ping: invalid number of pings.\n");exit(2); }
        break;
      case 'i':
        delay=atoi(optarg);
        if (delay<1 || delay>1000) { fprintf(stderr,"ping: invalid ping delay.\n");exit(2); }
        break;
      case 't':
        timeout=atoi(optarg);
        if (timeout<1 || timeout>100) { fprintf(stderr,"ping: invalid ping timeout.\n");exit(2); }
        break;
      default:
        usage();
    }
    
  if (argc-optind!=1) usage();
  
  min=timeout*1000000;
  
  hp=gethostbyname(argv[optind]);
  
  if (!hp) {
    fprintf(stderr,"ping: host not found.\n");
    exit(1);
  }
  
  nic=*(long*)hp->h_addr;
  
  sprintf(ad,"%d.%d.%d.%d",nic & 0xff,nic>>8 & 0xff,nic>>16 & 0xff,
          nic>>24 & 0xff);
	  
  signal(SIGINT,stats);
  
  port=65535;

  if (count) 
    printf("Pinging %s (%s) - %d packets, delay %d sec(s), timeout %d sec(s).\n",hp->h_name,ad,count,delay,timeout);
  else
    printf("Pinging %s (%s) - delay %d sec(s), timeout %d sec(s).\n",hp->h_name,ad,delay,timeout);

  while (n++<=count || !count) {
    bcopy(hp->h_addr,(void*)&s.sin_addr,hp->h_length);
    s.sin_family=hp->h_addrtype;
    s.sin_port=htons(port);
    if ((suck=socket(AF_INET,SOCK_STREAM,0))<0) xerr("ping");
    siga.sa_handler=alrm;
    to=tamto=skew=siga.sa_flags=0;
    sigaction(SIGALRM,&siga,NULL);
    alarm(timeout);
    gettimeofday(&tv,NULL);
    if ((skew=!connect(suck,(struct sockaddr *)&s,sizeof(s))) || 1) {
      if (to || errno==EINTR) printf("No ping reply from %s within %d second(s)...\n",ad,tamto=timeout);
        else if (errno && errno!=ECONNREFUSED) xerr("ping");
    }
    gettimeofday(&tv2,NULL);
    if (!tamto) {
      nic=(tv2.tv_sec-tv.tv_sec)*1000000+tv2.tv_usec-tv.tv_usec;
      if (nic/1000>=0.1) printf("Ping reply from %s: seq=%d, time=%.1f ms%s\n",ad,n-1,(float)nic/1000,skew?" (skew!)":"");
        else printf("Ping reply from %s: seq=%d, time<0.1 ms (%d usecs)%s\n",ad,n-1,nic,skew?" (skew!)":"");
      got++;
      trip+=nic;
      port-=skew;
      if (nic>max) max=nic;
      if (nic<min) min=nic;
      
    } else port-=1;
    shutdown(suck,2);
    close(suck);
    usleep(delay*1000000);
  }
  n--;
  stats(0);
  return(0);
}

